from __future__ import absolute_import

import logging
import os
import random
import re
import string

import azure.batch.models as batchmodels
import datetime
import time
import uuid
from azure.batch import BatchServiceClient
from azure.common.credentials import ServicePrincipalCredentials
from azure.profiles import KnownProfiles

from .blobclient import BlobStorageClient

KnownProfiles.default.use(KnownProfiles.latest)


class AzureBatchPoolError(Exception):
    pass


class AzureLoggingFilter(logging.Filter):
    def __init__(self):
        self.filter_keywords = [
            'azure', 'adal-python', 'urllib3', 'msrest', 'msal'
        ]

    def filter(self, rec):
        logname = rec.name.split('.')[0]
        if logname in self.filter_keywords:
            return False
        return True


class BatchClient(object):
    def __init__(
            self, client_id, tenant_id, secret_key, batch_account_url,
            config, storage_account_token=None
    ):
        """
        abstraction around all batch related methods
        :param client_id: Azure AD client id (application id)
        :type client_id: str
        :param tenant_id: Azure AD tenant id
        :type tenant_id: str
        :param secret_key: secret key for the app
        :type secret_key: str
        :param batch_account_url: azure batch account URL
        :type batch_account_url: str
        :param config: configuration file data from submit config
        :type config: dict
        :param storage_account_token: storage account key
        :type storage_account_token:  str
        """
        self.config = config

        self.run_id = self.__random_string(8)

        self.logger = self.__get_logger(add_azure_filter=True)

        self.logger.info('creating blob client')
        self.blob_client = BlobStorageClient(
            config['pypeliner_storage_account'],
            client_id, tenant_id, secret_key,
            storage_account_token=storage_account_token,
        )

        self.logger.info('creating batch client')
        self.credentials = ServicePrincipalCredentials(
            client_id=client_id,
            secret=secret_key,
            tenant=tenant_id,
            resource="https://batch.core.windows.net/"
        )
        self.batch_client = BatchServiceClient(
            self.credentials,
            batch_account_url
        )

        self.container_name = self.config['storage_container_name']
        self.blob_client.create_container(container_name=self.container_name)

        self.active_pools = set()
        self.active_jobs = set()

    def __get_vm_image_and_node_agent_sku(self, pool_config):
        """Select the latest verified image that Azure Batch supports given
        a publisher, offer and sku (starts with filter).
        Get the node agent SKU and image reference for the virtual machine
        configuration.
        For more information about the virtual machine configuration, see:
        https://azure.microsoft.com/documentation/articles/batch-linux-nodes/
        :param dict pool_config: vm configuration
        :rtype: tuple
        :return: (node agent sku id to use, vm image ref to use)
        """
        if pool_config.get("node_resource_id", None):
            image_ref_to_use = batchmodels.ImageReference(
                virtual_machine_image_id=pool_config["node_resource_id"])
            sku_to_use = pool_config['node_os_sku']
        else:
            publisher = pool_config['node_os_publisher']
            offer = pool_config['node_os_offer']
            sku_starts_with = pool_config['node_os_sku']

            # get verified vm image list and node agent sku ids from service
            node_agent_skus = self.batch_client.account.list_node_agent_skus()

            # pick the latest supported sku
            for sku in node_agent_skus:
                for image_ref in sorted(sku.verified_image_references, key=lambda item: item.sku):
                    if not image_ref.publisher.lower() == publisher.lower():
                        continue
                    if not image_ref.offer.lower() == offer.lower():
                        continue
                    if not image_ref.sku.startswith(sku_starts_with):
                        continue

                    sku_to_use = sku.id
                    image_ref_to_use = image_ref
                    break

        return sku_to_use, image_ref_to_use

    def __print_batch_exception(self, batch_exception):
        """
        Prints the contents of the specified Batch exception.

        :param batch_exception:
        """
        print('-------------------------------------------')
        print('Exception encountered:')
        if (batch_exception.error and batch_exception.error.message and
                batch_exception.error.message.value):
            print(batch_exception.error.message.value)
            if batch_exception.error.values:
                print()
                for msg in batch_exception.error.values:
                    print('{}:\t{}'.format(msg.key, msg.value))
        print('-------------------------------------------')

    def __create_commands(self, commands_str):
        return [a for a in commands_str.split('\n') if len(a) > 0]

    def __wrap_commands_in_shell(self, ostype, commands):
        """Wrap commands in a shell

        :param list commands: list of commands to wrap
        :param str ostype: OS type, linux or windows
        :rtype: str
        :return: a shell wrapping commands
        """
        if not commands:
            commands = []
        if ostype.lower() == 'linux':
            return '/bin/bash -c \'set -e; set -o pipefail; {}; wait\''.format(
                '; '.join(commands))
        elif ostype.lower() == 'windows':
            return 'cmd.exe /c "{}"'.format('&'.join(commands))
        else:
            raise ValueError('unknown ostype: {}'.format(ostype))

    def pool_exists(self, pool_id):
        """
        check if pool exists already
        :param pool_id: pool name
        :type pool_id: str
        :return: True if exists, False otherwise
        :rtype: bool
        """
        return self.batch_client.pool.exists(pool_id)

    def get_job_list(self):
        """
        list all jobs in batch account
        :return: list of jobs
        :rtype: list
        """
        job_ids = set([job.id for job in self.batch_client.job.list()])
        return job_ids

    def get_task_list(self, job_id, task_filter=None, list_max_results=None):
        """
        return tasks, optionally can filter them as well
        :param job_id: lists tasks under this Job Id
        :type job_id: str
        :param task_filter: filter string for list query
        :type task_filter:  str
        :param list_max_results: number of results to return
        :type list_max_results: int or None
        :return: list of tasks
        :rtype: list
        """
        if task_filter:
            list_options = batchmodels.TaskListOptions(filter=task_filter)
        else:
            list_options = None

        return self.batch_client.task.list(
            job_id, task_list_options=list_options, list_max_results=list_max_results
        )

    def create_job(self, pool_id, job_id):
        """
        Creates a job with the specified ID, associated with the specified pool.

        :param str job_id: The ID for the job.
        :param str pool_id: The ID for the pool.
        """

        if job_id in self.active_jobs or job_id in self.get_job_list():
            return

        self.logger.info("creating job {}".format(job_id))

        job = batchmodels.JobAddParameter(
            id=job_id,
            pool_info=batchmodels.PoolInformation(pool_id=pool_id))

        try:
            self.batch_client.job.add(job)
        except batchmodels.BatchErrorException as err:
            self.__print_batch_exception(err)
            raise

        for task in self.get_task_list(job_id):
            self.batch_client.task.delete(job_id, task.id)

        self.active_jobs.add(job_id)

    def create_pool(self, pool_id):
        """
        Creates a pool of compute nodes with the specified OS settings.

        :param str pool_id: An ID for the new pool.
        :param dict config: Configuration details.
        """
        if pool_id in self.active_pools or self.pool_exists(pool_id):
            return

        self.logger.info("creating pool {}".format(pool_id))

        pool_config = self.config['pools'][pool_id]

        sku_to_use, image_ref_to_use = self.__get_vm_image_and_node_agent_sku(
            pool_config
        )

        start_vm_commands = None
        if pool_config.get('create_vm_commands', None):
            start_vm_commands = self.__create_commands(pool_config['create_vm_commands'])

        user = batchmodels.AutoUserSpecification(
            scope=batchmodels.AutoUserScope.pool,
            elevation_level=batchmodels.ElevationLevel.admin)

        vm_configuration = batchmodels.VirtualMachineConfiguration(
            image_reference=image_ref_to_use,
            node_agent_sku_id=sku_to_use,
        )

        vm_start_task = batchmodels.StartTask(
            command_line=self.__wrap_commands_in_shell('linux', start_vm_commands),
            user_identity=batchmodels.UserIdentity(auto_user=user),
            wait_for_success=True)

        new_pool = batchmodels.PoolAddParameter(
            id=pool_id,
            virtual_machine_configuration=vm_configuration,
            vm_size=pool_config['pool_vm_size'],
            enable_auto_scale=True,
            auto_scale_formula=pool_config['auto_scale_formula'],
            auto_scale_evaluation_interval=datetime.timedelta(minutes=5),
            start_task=vm_start_task,
            max_tasks_per_node=pool_config['max_tasks_per_node'],
        )

        try:
            self.batch_client.pool.add(new_pool)
        except batchmodels.BatchErrorException as err:
            self.__print_batch_exception(err)
            raise

        self.active_pools.add(pool_id)

    def __get_logger(self, add_azure_filter=False):
        """
        generate a logger, optionally add filter to remove all azure sdk logs
        :param add_azure_filter: set the filter if set
        :type add_azure_filter: bool
        :return: python logger object
        :rtype: logging.Logger
        """
        logger = logging.getLogger('pypeliner.execqueue.azure_batch')
        if add_azure_filter:
            for handler in logging.root.handlers:
                handler.addFilter(AzureLoggingFilter())
        return logger

    def __random_string(self, length):
        return ''.join(random.choice(string.ascii_lowercase) for _ in range(length))

    def __wait_for_job_deletion(self, job_id):
        """
        Wait until the job is deleted. avoids issues when running
        multiple pipelines with same run id serially.
        :param str job_id: job identifier
        """

        while True:
            try:
                self.batch_client.job.get(job_id)
            except batchmodels.BatchErrorException:
                break
            time.sleep(30)

    def __wait_for_pool_deletion(self, pool_id):
        """
        Wait until the job is deleted. avoids issues when running
        multiple pipelines with same run id serially.
        :param str pool_id: job identifier
        """

        while True:
            try:
                self.batch_client.pool.get(pool_id)
            except batchmodels.BatchErrorException:
                break
            time.sleep(30)

    def __get_biggest_pool(self):
        """
        try to find the pool with the biggest node size
        :return: pool name
        :rtype: str
        """
        biggestpool = None

        for poolid, poolinfo in self.config['pools'].items():
            memory = poolinfo['mem_per_task']
            cpus = poolinfo['cpus_per_task']
            disk = poolinfo['disk_per_task']
            score = memory + (cpus * 8) + disk / 10

            if not biggestpool:
                biggestpool = (score, poolid)
            else:
                if score > biggestpool[0]:
                    biggestpool = (score, poolid)

        return biggestpool[1]

    def find_pool(self, ctx):
        """
        choose the best pool based on job context
        :param ctx: dict, job context dict
        :return: str, name of pool that fits job context
        """
        if ctx.get('pool_id', None):
            return ctx['pool_id']
        memory_req = ctx.get('mem', 4)
        cpus_req = ctx.get('ncpus', 1)
        disk_req = ctx.get('disk', 8)
        dedicated_req = ctx.get('dedicated', False)

        # at the ctx level, memory is normally specified per cpu
        # in batch yaml memory_per_task is less confusing than memory_per_cpu
        memory_req_per_task = memory_req * cpus_req

        pools = []

        for poolid, poolinfo in self.config['pools'].items():
            memory = poolinfo['mem_per_task']
            cpus = poolinfo['cpus_per_task']
            disk = poolinfo['disk_per_task']
            dedicated = poolinfo.get('dedicated', None)

            # dont submit dedicated jobs to low priority nodes
            if dedicated_req and not dedicated:
                continue

            if memory_req_per_task <= memory and cpus_req <= cpus and disk_req <= disk:
                # giving a bigger weight to cpu since cpu number is always smaller
                # smaller weight to disk. score is an approximation of amount of wasted
                # resources
                distance = (memory - memory_req_per_task) + ((cpus - cpus_req) * 8) + (disk - disk_req) / 10
                pools.append((distance, poolid))

        if not pools:
            biggestpool = self.__get_biggest_pool()
            warn_str = \
                "Could not find a pool to satisfy job requirements. " \
                "requested mem:{} cpus:{} dedicated: {}." \
                " submitting to {}".format(
                    memory_req_per_task, cpus_req, dedicated_req, biggestpool)
            self.logger.warn(warn_str)
            return biggestpool

        # choose best fit
        pools = sorted(pools, key=lambda tup: tup[0])

        return pools[0][1]

    @property
    def compute_start_commands(self):
        compute_start_commands = {
            id: pool['compute_start_commands'] for id, pool in self.config['pools'].items()
        }
        return compute_start_commands

    @property
    def compute_finish_commands(self):
        compute_finish_commands = {
            id: pool['compute_finish_commands'] for id, pool in self.config['pools'].items()
        }
        return compute_finish_commands

    def get_node_id_for_task(self, job_id, task_id):
        """
        find the node that task was scheduled on
        :param job_id: batch job id
        :type job_id: str
        :param task_id: batch task id
        :type task_id: str
        :return: name of node (azure affinity id)
        :rtype: str
        """
        node_info = self.batch_client.task.get(
            job_id,
            task_id).node_info
        if not node_info:
            return
        hostname = node_info.node_id
        return hostname

    def get_job_id(self, pool_id):
        return 'pypeliner_{}_{}'.format(pool_id, self.run_id)

    def delete_pool(self, pool_id, ignore_exception=False):
        """
        try to delete a pool
        :param pool_id: batch pool id
        :type pool_id: str
        :param ignore_exception: dont raise any errors if set
        :type ignore_exception: bool
        """
        try:
            self.batch_client.pool.delete(pool_id)
        except Exception as e:
            if ignore_exception:
                print(e)
            else:
                raise e

    def delete_job(self, job_id, ignore_exception=False):
        """
        try to delete a job
        :param job_id:
        :type job_id: str
        :param ignore_exception: dont raise any errors if set
        :type ignore_exception: bool
        """
        try:
            self.batch_client.job.delete(job_id)
        except Exception as e:
            if ignore_exception:
                print(e)
            else:
                raise e

    def delete_jobs(self, ignore_exception=False, wait=False):
        """
        delete all jobs that were created by this instance
        :param ignore_exception: dont raise any errors if set
        :type ignore_exception: bool
        :param wait: wait until all jobs are deleted
        :type wait: bool
        """
        for job_id in self.active_jobs:
            self.delete_job(job_id, ignore_exception=ignore_exception)

        if wait:
            for job_id in self.active_jobs:
                self.__wait_for_job_deletion(job_id)

    def delete_pools(self, ignore_exception=False, wait=False):
        """
        delete all pools that were created by this instance
        :param ignore_exception: dont raise any errors if set
        :type ignore_exception: bool
        :param wait: wait until all jobs are deleted
        :type wait: bool
        """
        for pool_id in self.active_pools:
            self.delete_pool(pool_id, ignore_exception=ignore_exception)

        if wait:
            for pool_id in self.active_pools:
                self.__wait_for_pool_deletion(pool_id)

    @property
    def healthy_states(self):
        healthy_state_list = [
            "idle", "running", "preempted",
            "waitingforstarttask", "leavingpool", 'starting'
        ]
        return healthy_state_list

    def verify_node_status(self, node_id, pool_id):
        """
        check if the node is in working state, if not disable it.
        :param str node_id: compute node id
        :param str pool_id: batch pool id
        """

        status = self.__get_node_status(pool_id, node_id)

        # dont send jobs to failed nodes
        if status not in self.healthy_states and not status == "offline":
            self.logger.warning(
                'node {} is in {} state and will be disabled'.format(
                    node_id, status)
            )
            self.__disable_node(pool_id, node_id)

    def __disable_node(self, pool_id, node_id):
        self.batch_client.compute_node.disable_scheduling(pool_id, node_id)

    def __list_compute_nodes_in_pool(self, pool_id):
        for node in self.batch_client.compute_node.list(pool_id):
            yield node.id

    def __get_node_status(self, pool_id, node_id):
        try:
            status = self.batch_client.compute_node.get(
                pool_id,
                node_id).state.value
        except batchmodels.BatchErrorException:
            self.logger.warning("Couldn't get status for node {} ".format(node_id))
            status = 'error'
        return status

    def check_pool_for_failed_nodes(self, pool_id):
        """
        Collects node state information from the batch pool
        :param batch_client: batch service client
        :param str pool_id: ID of pool to monitor
        :param logger: logger object to issue warnings
        """

        healthy_states = ["idle", "running", "preempted",
                          "waitingforstarttask", "leavingpool", 'starting']

        for node_id in self.__list_compute_nodes_in_pool(pool_id):
            status = self.__get_node_status(pool_id, node_id)

            # node states are inconsistent with lower and upper cases
            # stop looking when at least one working node exists
            if status.lower() in healthy_states:
                return

        self.logger.warning(
            "Couldn't find any working nodes in pool. Please verify pool status")

    def download_blobs_from_container(
            self, directory_path, prefix
    ):

        """
        Downloads all blobs from the specified Azure Blob storage container.
        :param directory_path: The local directory to which to download the files.
        :type directory_path: str
        :param prefix: removes this prefix from blob name
        :type prefix:str
        """
        container_blobs = self.blob_client.list_blobs(container_name=self.container_name, prefix=prefix)

        for blob in container_blobs:
            destination_filename = blob.name

            # Remove prefix
            assert destination_filename.startswith(prefix)
            destination_filename = destination_filename[(len(prefix)):]
            destination_filename = destination_filename.strip('/')

            destination_file_path = os.path.join(
                directory_path,
                destination_filename)

            try:
                os.makedirs(os.path.dirname(destination_file_path))
            except:
                pass

            self.blob_client.download_to_path(
                destination_file_path, container_name=self.container_name, blob_name=blob.name
            )

    def create_blob_resource(
            self, file_path, blob_name, job_file_path):
        """
        Uploads a local file to an Azure Blob storage container for use as
        an azure batch resource.

        :param file_path: The local path to the file.
        :type file_path: str
        :param blob_name: path to upload to
        :type blob_name: str
        :param job_file_path: filepath for batch resource
        :type job_file_path: str
        :rtype: `azure.batch.models.ResourceFile`
        :return: A ResourceFile initialized with a SAS URL appropriate for Batch
        """

        self.blob_client.upload_from_file(
            file_path, container_name=self.container_name, blob_name=blob_name
        )

        sas_url = self.blob_client.generate_blob_url(
            container_name=self.container_name, blob_name=blob_name, add_sas=True
        )

        return batchmodels.ResourceFile(
            file_path=job_file_path,
            http_url=sas_url
        )

    def add_task(
            self, job_id, task_id, input_file, job_script_file,
            output_blob_prefix
    ):
        """
        Adds a task for each input file in the collection to the specified job.

        :param job_id: The ID of the job to which to add the tasks.
        :type job_id: str
        :param task_id: The ID of the task.
        :type task_id: str
        :param input_file: delegator input file
        :type input_file: str
        :param job_script_file: job.sh file
        :type job_script_file: str
        :param output_blob_prefix: prefix where output files will be uploaded to
        :type output_blob_prefix: str
        """
        commands = ['/bin/bash job.sh']

        resource_files = [input_file, job_script_file]

        # Obtain a shared access signature that provides write access to the output
        # container to which the tasks will upload their output.
        output_sas_url = self.blob_client.generate_container_url(
            container_name=self.container_name, add_sas=True, sas_permissions='create|write')

        def create_output_file(filename, blobname):
            container = batchmodels.OutputFileBlobContainerDestination(
                container_url=output_sas_url, path=blobname)
            destination = batchmodels.OutputFileDestination(container=container)
            upload_options = batchmodels.OutputFileUploadOptions(upload_condition='taskCompletion')
            return batchmodels.OutputFile(file_pattern=filename, destination=destination, upload_options=upload_options)

        task = batchmodels.TaskAddParameter(
            id=task_id,
            command_line=self.__wrap_commands_in_shell('linux', commands),
            resource_files=resource_files,
            output_files=[
                create_output_file(
                    '../stdout.txt',
                    os.path.join(
                        output_blob_prefix,
                        'stdout.txt')),
                create_output_file(
                    '../stderr.txt',
                    os.path.join(
                        output_blob_prefix,
                        'stderr.txt')),
                create_output_file(
                    'job_result.pickle',
                    os.path.join(
                        output_blob_prefix,
                        'job_result.pickle')),
            ]
        )

        self.batch_client.task.add(job_id, task)

    def get_task_id(self, name, azure_task_name_max_len=64):
        """
        generate uniq string, fill in the leftover space
        with job information
        :param name: str, pypeliner's internal task name
        :param azure_task_name_max_len: azure maximum allowed task name length
        :return str, unique azure compatible task id
        """

        uniq_string = str(uuid.uuid1())

        leftover_space = azure_task_name_max_len - len(uniq_string) - 1

        # last few characters are more useful for identifying tasks
        name = name[-leftover_space:]

        name = '-'.join(re.split('[:/?]', name))

        name = "{}-{}".format(uniq_string, name)

        return name

    def get_pool_id_from_job(self, job_id):
        """
        get pool id from job id
        :param str job_id: azure batch job id
        :return str: azure batch pool id
        """
        return self.batch_client.job.get(job_id).pool_info.pool_id

    def check_for_missing_node(self, job_id, task_id):
        node = self.get_node_id_for_task(job_id, task_id)
        pool = self.get_pool_id_from_job(job_id)

        if not node:
            self.logger.debug(
                'task {} under job {} and in pool {} '
                'has not been scheduled on a node yet!'.format(
                    task_id, job_id, pool
                )
            )
        else:
            status = self.__get_node_status(pool, node)

            if status not in self.healthy_states:
                logging.debug(
                    'task {} under job {} and in pool {} was scheduled '
                    'on node {} which is now in {} state'.format(
                        task_id, job_id, pool, node, status
                    )
                )



import argparse
import os
import logging
import yaml

import pypeliner.execqueue.factory
import pypeliner.helpers


class BasicJob():
    def __init__(self):
        self.success = False
        self.started = False
        self.ctx = {}
        self.log_records = []
        self.version = pypeliner.__version__
    def __call__(self):
        self.started = True
        self.success = True


def run_basic(exec_queue, base_temps_dir, ctx):

    assert exec_queue.empty

    job = test_queue.BasicJob()
    temps_dir = os.path.join(base_temps_dir, 'basic')
    pypeliner.helpers.makedirs(temps_dir)
    exec_queue.send(ctx, 'basic', job, temps_dir)

    assert exec_queue.length == 1
    assert not exec_queue.empty

    name = exec_queue.wait()

    assert name == 'basic'

    recieved = exec_queue.receive(name)

    assert recieved.success == True

    print ('success')


if __name__ == '__main__':
    from pypeliner.tests import test_queue

    argparser = argparse.ArgumentParser()

    argparser.add_argument('submit',
        help='Execution queue to test')

    argparser.add_argument('--nativespec', default=None,
        help='Native spec if needed')

    argparser.add_argument('--configyaml', default=None,
        help='Optional config yaml')

    args = vars(argparser.parse_args())

    logging.basicConfig(level=logging.DEBUG)
    console = logging.StreamHandler()
    console.setLevel(logging.DEBUG)
    console.setFormatter(pypeliner.helpers.MultiLineFormatter('%(asctime)s - %(name)s - %(levelname)s - '))
    logging.getLogger('').addHandler(console)

    script_directory = os.path.dirname(os.path.abspath(__file__))
    base_temps_dir = os.path.join(script_directory, 'queue_test')

    exec_queue = pypeliner.execqueue.factory.create(args['submit'], [test_queue], native_spec=args['nativespec'])

    ctx = {'mem': 1}
    if args.get('configyaml') is not None:
        ctx = yaml.safe_load(open(args['configyaml']), Loader=yaml.FullLoader)

    with exec_queue:
        run_basic(exec_queue, base_temps_dir, ctx)




import errno
import hashlib
import importlib
import json
import logging
import os
import random
import shutil
import stat
import time
from collections import deque
from functools import wraps

from pypeliner import _pypeliner_internal_global_state


def running_in_singularity():
    if 'SINGULARITY_NAME' in os.environ:
        return True
    else:
        return False

def running_in_docker():
    """
    returns True if python is running inside docker container
    cgroup file is normally readable by all users
    """
    with open('/proc/self/cgroup', 'r') as procfile:
        for line in procfile:
            fields = line.strip().split('/')
            if 'docker' in fields:
                return True
    return False


def log_event(info, name=None, extras=None, logger=None, level='info'):
    if logger:
        name = name if name else 'pypeliner.helpers'
        logger = logging.getLogger(name)

    if isinstance(info, dict):
        # add primary info to json field
        extras.update(info)
        info = [i for keyvalue in info.iteritems() for i in keyvalue]

    if isinstance(info, list) or isinstance(info, tuple):
        parsed_info = []
        for val in info:
            if isinstance(val, list) or isinstance(val, tuple):
                extras[val[0]] = val[1]
                parsed_info.extend(list(val))
            else:
                parsed_info.append(val)
        info = ' '.join(map(str, parsed_info))

    if level == 'warn':
        logger.warning(info, extra=extras)
    elif level == 'err':
        logger.error(info, extra=extras)
    elif level == 'info':
        logger.info(info, extra=extras)
    elif level == 'debug':
        logger.debug(info, extra=extras)
    else:
        raise NotImplementedError('Cannot handle {} logging level'.format(level))


class GlobalState(object):
    """
    add the specified variable and value to a pypeliner wide
    key value store
    :param variablename: key name
    :param value: key value
    """

    @staticmethod
    def get(variablename, default_value=None):
        return _pypeliner_internal_global_state.get(variablename, default_value)

    @staticmethod
    def get_all():
        return _pypeliner_internal_global_state

    @staticmethod
    def set(variablename, value):
        _pypeliner_internal_global_state[variablename] = value

    @staticmethod
    def update_all(globals_dict):
        _pypeliner_internal_global_state.update(globals_dict)


class Backoff(object):
    """
    wrapper for functions that fail but require retries until it succeeds
    """

    def __init__(self, exception_type=Exception, max_backoff=3600, backoff_time=1, randomize=False,
                 num_retries=None, backoff="exponential", step_size=2):
        self.func = None

        if backoff not in ["exponential", "linear", "fixed"]:
            raise Exception(
                "Currently supports only exponential, linear and fixed backoff")

        self.exception_type = exception_type

        self.max_backoff = max_backoff
        self.backoff_time = backoff_time

        self.randomize = randomize

        self.num_retries = num_retries

        self.backoff = backoff

        self.step_size = step_size

        self.elapsed_time = 0

    def __call__(self, func):
        self.func = func

        @wraps(func)
        def wrapped(*args, **kwargs):
            return self._run_with_exponential_backoff(*args, **kwargs)

        return wrapped

    def _run_with_exponential_backoff(self, *args, **kwargs):
        """
        keep running the function until we go over the
        max wait time or num retries
        """

        retry_no = 0

        while True:

            if self.elapsed_time >= self.max_backoff:
                break

            if self.num_retries and retry_no > self.num_retries:
                break

            try:
                result = self.func(*args, **kwargs)
            except self.exception_type as exc:
                self._update_backoff_time()
                logging.getLogger("pypeliner.helpers").warn(
                    "error {} caught, retrying after {} seconds".format(
                        str(exc), self.backoff_time)
                )
                retry_no += 1
                time.sleep(self.backoff_time)
            except Exception:
                raise
            else:
                return result

    def _update_backoff_time(self):
        """
        update the backoff time
        """

        if self.backoff == "exponential":
            self.backoff_time = self.step_size * (self.backoff_time or 1)

        elif self.backoff == "linear":
            self.backoff_time += self.step_size

        if self.randomize:
            lower_bound = int(0.9 * self.backoff_time)
            upper_bound = int(1.1 * self.backoff_time)
            self.backoff_time = random.randint(lower_bound, upper_bound)

        if self.elapsed_time + self.backoff_time > self.max_backoff:
            self.backoff_time = self.max_backoff - self.elapsed_time

        self.elapsed_time += self.backoff_time


def import_function(import_string):
    module, funcname = import_string.rsplit('.', 1)

    mod = importlib.import_module(module)
    met = getattr(mod, funcname)

    return met


def pop_if(L, pred):
    for idx, item in enumerate(L):
        if pred(*item):
            return L.pop(idx)
    raise IndexError()


def abspath(path):
    if path.endswith('/'):
        return os.path.abspath(path) + '/'
    else:
        return os.path.abspath(path)


class MultiLineFormatter(logging.Formatter):
    def format(self, record):
        header = logging.Formatter.format(self, record)
        return header + record.message.rstrip('\n').replace('\n', '\n\t')


class JsonFormatter(logging.Formatter):
    def format(self, record):
        return json.dumps(vars(record))


def which(name):
    if os.environ.get('PATH', None) is not None:
        for p in os.environ.get('PATH', '').split(os.pathsep):
            p = os.path.join(p, name)
            if os.access(p, os.X_OK):
                return p
    raise EnvironmentError('unable to find ' + name + ' in the system path')


def set_executable(filename):
    mode = os.stat(filename).st_mode
    mode |= stat.S_IXUSR
    os.chmod(filename, stat.S_IMODE(mode))


def md5_file(filename, block_size=8192):
    md5 = hashlib.md5()
    with open(filename, 'rb') as f:
        for chunk in iter(lambda: f.read(block_size), b''):
            md5.update(chunk)
    return md5.digest()


def overwrite_if_different(new_filename, existing_filename):
    do_copy = True
    try:
        do_copy = md5_file(existing_filename) != md5_file(new_filename)
    except IOError:
        pass
    if do_copy:
        os.rename(new_filename, existing_filename)


def makedirs(dirname):
    dirname = abspath(dirname)
    try:
        os.makedirs(dirname)
    except OSError as e:
        if e.errno != errno.EEXIST:
            raise
    assert os.path.isdir(dirname)


def saferemove(filename):
    try:
        os.remove(filename)
    except OSError:
        pass


def symlink(source, link_name):
    source = os.path.abspath(source)
    try:
        os.remove(link_name)
    except OSError as e:
        if e.errno != errno.ENOENT:
            raise
    os.symlink(source, link_name)


def touch(filename, times=None):
    with open(filename, 'a'):
        os.utime(filename, times)


def removefiledir(filename):
    saferemove(filename)
    shutil.rmtree(filename, ignore_errors=True)


class RemoteLogHandler(logging.Handler):
    def __init__(self, logs):
        logging.Handler.__init__(self)
        self.logs = logs

    def emit(self, log_record):
        self.logs.append(log_record)


class RemoteLogger(object):
    def __init__(self):
        self._logs = deque(maxlen=1000)
        self._handler = RemoteLogHandler(self._logs)

    @property
    def log_records(self):
        return self._logs

    @property
    def log_handler(self):
        return self._handler



import os

import datetime
import pypeliner.flyweight
import pypeliner.helpers
import pypeliner.storage
import time
from pypeliner.sqlitedb import SqliteDb

from .blobclient import BlobStorageClient, BlobMissing


class AzureBlob(object):
    def __init__(self, storage, filename, **kwargs):
        self.storage = storage
        store_dir = kwargs.get("store_dir")
        extension = kwargs.get("extension")
        direct_write = kwargs.get("direct_write", False)

        self.filename = filename
        self.blob_name = filename
        self.write_filename = filename + ('.tmp', '')[direct_write]

        if extension:
            self.filename += extension
            self.blob_name += extension
            self.write_filename += extension

        if store_dir and not direct_write:
            self.filename = os.path.join(store_dir, self.filename)
            self.write_filename = os.path.join(store_dir, self.write_filename)

        self.createtime_cache = kwargs.get('createtime_cache')
        self.createtime_save = kwargs.get('createtime_save')
        self.exists_cache = kwargs.get('exists_cache')

    def allocate(self):
        pypeliner.helpers.makedirs(os.path.dirname(self.filename))

    def push(self):
        try:
            os.rename(self.write_filename, self.filename)
        except OSError:
            raise pypeliner.storage.OutputMissingException(self.write_filename)

        create_time = datetime.datetime.fromtimestamp(
            os.path.getmtime(self.filename)
        )
        create_time = create_time.strftime('%Y/%m/%d-%H:%M:%S')

        self.storage.push(self.blob_name, self.filename, create_time)
        self.createtime_cache.set(create_time)
        self.createtime_save.set(create_time)
        self.exists_cache.set(True)

    def pull(self):
        self.storage.pull(self.blob_name, self.filename)

    def get_exists(self):
        exists = self.exists_cache.get()
        if exists is None:
            createtime = self.get_createtime()

            if createtime is not None and createtime != 'missing':
                exists = True
            else:
                exists = False
            self.exists_cache.set(exists)
        return exists

    def get_createtime(self):
        sentinel_only = pypeliner.helpers.GlobalState.get('sentinel_only')
        if sentinel_only:
            createtime = self.createtime_save.get()
        else:
            createtime = self.createtime_cache.get()

        if createtime is None:
            try:
                createtime = self.storage.retrieve_blob_createtime(self.blob_name)
            except BlobMissing:
                createtime = 'missing'
            self.createtime_cache.set(createtime)
            self.createtime_save.set(createtime)
        if createtime == 'missing':
            return None
        createtime = datetime.datetime.strptime(createtime, '%Y/%m/%d-%H:%M:%S')
        return time.mktime(createtime.timetuple())

    def touch(self):
        createtime = datetime.datetime.now().strftime('%Y/%m/%d-%H:%M:%S')
        self.exists_cache.set(True)
        self.storage.update_blob_createtime(self.blob_name, createtime)
        self.createtime_cache.set(createtime)
        self.createtime_save.set(createtime)

    def delete(self):
        self.storage.delete_blob(self.blob_name)
        self.createtime_cache.set(None)


class AzureBlobStorage(object):
    def __init__(self, **kwargs):
        self.client_id = os.environ["CLIENT_ID"]
        self.secret_key = os.environ["SECRET_KEY"]
        self.tenant_id = os.environ["TENANT_ID"]
        self.subscription_id = os.environ["SUBSCRIPTION_ID"]
        self.blob_client = None

        metadata_prefix = kwargs.get('metadata_prefix')
        createtime_shelf_filename = metadata_prefix + 'createtimes.db'
        self.cached_createtimes = pypeliner.flyweight.FlyweightState()
        pypeliner.helpers.makedirs(os.path.dirname(createtime_shelf_filename))
        self.saved_createtimes = pypeliner.flyweight.FlyweightState(
            state_container=SqliteDb(createtime_shelf_filename))
        self.cached_exists = pypeliner.flyweight.FlyweightState()

    def connect(self, storage_account_name):
        # dont need to regenerate client if already exists
        client = getattr(self, "blob_client", None)
        if client and client.account_name == storage_account_name:
            return

        storage_keys = pypeliner.helpers.GlobalState.get('azure_storage_keys', {})
        account_token = storage_keys.get(storage_account_name)

        self.blob_client = BlobStorageClient(
            storage_account_name, self.client_id, self.tenant_id,
            self.secret_key,
            storage_account_token=account_token,
        )

    def create_createtime_cache(self, blob_name):
        return self.cached_createtimes.create_flyweight(blob_name)

    def __enter__(self):
        self.cached_exists.__enter__()
        self.cached_createtimes.__enter__()
        self.saved_createtimes.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.cached_exists.__exit__(exc_type, exc_value, traceback)
        self.cached_createtimes.__exit__(exc_type, exc_value, traceback)
        self.saved_createtimes.__exit__(exc_type, exc_value, traceback)

    def __getstate__(self):
        return (
            self.cached_createtimes, self.saved_createtimes, self.cached_exists,
            self.client_id, self.secret_key, self.tenant_id, self.subscription_id
        )

    def __setstate__(self, state):
        [self.cached_createtimes, self.saved_createtimes, self.cached_exists,
         self.client_id, self.secret_key, self.tenant_id, self.subscription_id] = state

    def create_store(self, filename, **kwargs):
        kwargs['createtime_cache'] = self.cached_createtimes.create_flyweight(filename)
        kwargs['createtime_save'] = self.saved_createtimes.create_flyweight(filename)
        kwargs['exists_cache'] = self.cached_exists.create_flyweight(filename)

        return AzureBlob(self, filename, **kwargs)

    def get_storage_account(self, filename):
        filename = filename.strip('/')
        filename = filename.split('/')
        storage_account = filename[0]
        return storage_account

    def push(self, blob_name, filename, createtime):
        self.connect(self.get_storage_account(blob_name))
        self.blob_client.upload_from_file(
            filename,
            metadata={'create_time': createtime},
            blob_uri=blob_name
        )

    def pull(self, blob_name, filename):
        self.connect(self.get_storage_account(blob_name))
        self.blob_client.download_to_path(
            filename,
            blob_uri=blob_name
        )

    def retrieve_blob_createtime(self, blob_name):
        self.connect(self.get_storage_account(blob_name))
        createtime = self.blob_client.get_metadata(
            'create_time', blob_uri=blob_name
        )
        if not createtime:
            return self.blob_client.get_last_modified(blob_uri=blob_name)
        else:
            return createtime

    def update_blob_createtime(self, blob_name, createtime):
        self.connect(self.get_storage_account(blob_name))
        self.blob_client.set_metadata(
            'create_time', createtime, blob_uri=blob_name
        )

    def delete_blob(self, blob_name):
        self.connect(self.get_storage_account(blob_name))
        self.blob_client.delete_blob(blob_uri=blob_name)



import importlib

from pypeliner.sqlitedb import SqliteDb
import os

import pypeliner.flyweight
import pypeliner.helpers


class InputMissingException(Exception):
    def __init__(self, filename):
        self.filename = filename

    def __str__(self):
        return 'expected input {} missing'.format(self.filename)


class OutputMissingException(Exception):
    def __init__(self, filename):
        self.filename = filename

    def __str__(self):
        return 'expected output {0} missing'.format(self.filename)


class RegularFile(object):
    def __init__(
            self, filename, exists_cache, createtime_cache, createtime_save,
            extension=None, direct_write=True, **kwargs
    ):
        self.filename = filename
        self.exists_cache = exists_cache
        self.createtime_cache = createtime_cache
        self.createtime_save = createtime_save
        self.write_filename = filename + ('.tmp', '')[direct_write]
        if extension is not None:
            self.filename = filename + extension
            self.write_filename = self.write_filename + extension

    def allocate(self):
        if not os.path.exists(self.filename):
            pypeliner.helpers.makedirs(os.path.dirname(self.filename))

    def push(self):
        try:
            os.rename(self.write_filename, self.filename)
        except OSError:
            raise OutputMissingException(self.write_filename)
        self.exists_cache.set(True)
        createtime = os.path.getmtime(self.filename)
        self.createtime_cache.set(createtime)
        self.createtime_save.set(createtime)

    def pull(self):
        if not self.get_exists():
            raise InputMissingException(self.filename)

    def get_exists(self):
        exists = self.exists_cache.get()
        if exists is None:
            exists = os.path.exists(self.filename)
            self.exists_cache.set(exists)
        return exists

    def get_createtime(self):
        # if sentinal only flag is set then pull create time from database too
        # to avoid hitting filesystem. this will cause issues if files
        # are deleted after pipeline run and before rerun
        sentinel_only = pypeliner.helpers.GlobalState.get('sentinel_only')
        if sentinel_only:
            createtime = self.createtime_save.get()
        else:
            createtime = self.createtime_cache.get()
        if createtime is None:
            if not self.get_exists():
                return None
            createtime = os.path.getmtime(self.filename)
            self.createtime_cache.set(createtime)
            self.createtime_save.set(createtime)
        if not isinstance(createtime, float):
            createtime = float(createtime)
        return createtime

    def touch(self):
        pypeliner.helpers.touch(self.filename)
        self.exists_cache.set(True)
        createtime = os.path.getmtime(self.filename)
        self.createtime_cache.set(createtime)
        self.createtime_save.set(createtime)

    def delete(self):
        raise Exception('cannot delete non-temporary files')


class RegularTempFile(RegularFile):
    def get_createtime(self):
        if super(RegularTempFile, self).get_exists():
            return float(super(RegularTempFile, self).get_createtime())
        createtime = self.createtime_save.get()
        if createtime and not isinstance(createtime, float):
            createtime = float(createtime)
        return createtime

    def delete(self):
        pypeliner.helpers.saferemove(self.filename)


class FileStorage(object):
    def __init__(self, metadata_prefix=None, **kwargs):
        createtime_shelf_filename = metadata_prefix + 'createtimes.db'
        pypeliner.helpers.makedirs(os.path.dirname(createtime_shelf_filename))
        self.cached_exists = pypeliner.flyweight.FlyweightState()
        self.cached_createtimes = pypeliner.flyweight.FlyweightState()
        self.saved_createtimes = pypeliner.flyweight.FlyweightState(
            state_container=SqliteDb(createtime_shelf_filename))

    def __enter__(self):
        self.cached_exists.__enter__()
        self.cached_createtimes.__enter__()
        self.saved_createtimes.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.cached_exists.__exit__(exc_type, exc_value, traceback)
        self.cached_createtimes.__exit__(exc_type, exc_value, traceback)
        self.saved_createtimes.__exit__(exc_type, exc_value, traceback)

    def _create_store(self, filename, factory, **kwargs):
        exists_cache = self.cached_exists.create_flyweight(filename)
        createtime_cache = self.cached_createtimes.create_flyweight(filename)
        createtime_save = self.saved_createtimes.create_flyweight(filename)
        return factory(filename, exists_cache, createtime_cache, createtime_save, **kwargs)

    def create_store(self, filename, is_temp=False, **kwargs):
        if is_temp:
            return self._create_store(filename, RegularTempFile, **kwargs)
        else:
            return self._create_store(filename, RegularFile, **kwargs)


def create(requested_storage, workflow_dir=None):
    if requested_storage is None:
        raise Exception('No storage specified')
    elif requested_storage == 'local':
        storage_name = 'pypeliner.storage.FileStorage'
    elif requested_storage == 'azureblob':
        storage_name = 'pypeliner.contrib.azure.blobstorage.AzureBlobStorage'
    elif requested_storage == 'awss3':
        storage_name = 'pypeliner.contrib.aws.objectstorage.AwsStorage'
    else:
        storage_name = requested_storage

    storage_class_name = storage_name.split('.')[-1]
    storage_module_name = storage_name[:-len(storage_class_name) - 1]

    storage_module = importlib.import_module(storage_module_name)
    storage_class = vars(storage_module)[storage_class_name]

    file_storage_prefix = os.path.join(workflow_dir, 'files_')

    storage = storage_class(metadata_prefix=file_storage_prefix)

    return storage



from setuptools import setup, find_packages

import versioneer

setup(
    name='pypeliner',
    packages=find_packages(),
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description='A library for creating informatic workflows or pipelines.',
    author='Andrew McPherson',
    author_email='andrew.mcpherson@gmail.com',
    url='http://bitbucket.org/dranew/pypeliner',
    download_url='https://bitbucket.org/dranew/pypeliner/get/v{}.tar.gz'.format(versioneer.get_version()),
    keywords=['scientific', 'framework'],
    classifiers=[],
    package_data={'pypeliner': ['tests/*.input']},
    entry_points={'console_scripts': ['pypeliner_delegate=pypeliner.delegator:main',
                                      'aws_fetch_run=pypeliner.contrib.aws.fetch_run:main'],
    },
    install_requires=[
        'dill',
        'networkx',
        'pyyaml',
        'six',
    ],
)



from __future__ import absolute_import

import os
import datetime
import time

import pypeliner.helpers
import pypeliner.storage
import pypeliner.flyweight

from .aws_storage import AwsSimpleStorageService

import pypeliner.contrib.aws.helpers as aws_helpers


class AwsObjectStorage(object):
    """
    :param storage: storage object
    :type storage: object
    :param filename: filename for the storage
    :type filename: str
    :param kwargs: store_dir, extension and direct_write
    :type kwargs: dict
    """
    def __init__(self, storage, filename, **kwargs):
        self.storage = storage
        if filename.startswith('/'):
            filename = filename[1:]
        store_dir = kwargs.get("store_dir")
        extension = kwargs.get("extension")
        direct_write = kwargs.get("direct_write", False)

        self.object_name = filename
        if self.object_name.startswith('/'):
            self.object_name = self.object_name[1:]
        if extension:
            self.object_name += extension
        self.write_filename = filename + ('.tmp', '')[direct_write]
        self.filename = filename

        self.createtime_cache = storage.create_createtime_cache(self.object_name)

        if extension is not None:
            self.filename = filename + extension
            self.write_filename = self.write_filename + extension
        if store_dir and not direct_write:
            self.filename = os.path.join(store_dir, self.filename)
            self.write_filename = os.path.join(store_dir, self.write_filename)

    def allocate(self):
        pypeliner.helpers.makedirs(os.path.dirname(self.filename))

    def push(self):
        try:
            os.rename(self.write_filename, self.filename)
        except OSError:
            raise pypeliner.storage.OutputMissingException(self.write_filename)
        create_time = datetime.datetime.fromtimestamp(os.path.getmtime(self.filename)).strftime('%Y/%m/%d-%H:%M:%S')
        self.storage.push(self.object_name, self.filename, create_time)
        self.createtime_cache.set(create_time)

    def pull(self):
        self.storage.pull(self.object_name, self.filename)

    def get_exists(self):
        createtime = self.get_createtime()
        return createtime is not None and createtime != 'missing'

    def get_createtime(self):
        createtime = self.createtime_cache.get()
        if createtime is None:
            if not self.storage.exists(self.object_name):
                createtime = 'missing'
            else:
                createtime = self.storage.retrieve_object_createtime(self.object_name)
            self.createtime_cache.set(createtime)
        if createtime == 'missing':
            return None
        createtime = datetime.datetime.strptime(createtime, '%Y/%m/%d-%H:%M:%S')
        return time.mktime(createtime.timetuple())

    def touch(self):
        createtime = datetime.datetime.now().strftime('%Y/%m/%d-%H:%M:%S')
        self.storage.update_object_createtime(self.object_name, createtime)
        self.createtime_cache.set(createtime)

    def delete(self):
        self.storage.delete_object(self.object_name)
        self.createtime_cache.set(None)


class AwsStorage(object):
    def __init__(self, **kwargs):
        aws_helpers.set_aws_logging_filters()
        self.cached_createtimes = pypeliner.flyweight.FlyweightState()
        self.object_client = AwsSimpleStorageService()

    def connect(self):
        """
        generate the s3 client if needed.
        :return:
        :rtype:
        """
        # dont need to regenerate client if already exists
        client = getattr(self, "object_client", None)
        if client:
            return
        self.object_client = AwsSimpleStorageService()

    def exists(self, object_name):
        """
        check if specified object exists
        :param object_name: name of object, should start with bucket name
        :type object_name: str
        :return: True if exists in S3
        :rtype: bool
        """
        bucket, object_name = self.unpack_path(object_name)
        self.connect()
        if not self.object_client.bucket_exists(bucket):
            return False
        if not self.object_client.key_exists(bucket, object_name):
            return False
        return True

    def create_createtime_cache(self, object_name):
        """
        cache the create time
        :param object_name: name of object, should start with bucket name
        :type object_name: str
        :return: flyweight cache object
        :rtype: pypeliner.flyweight.ReattachableFlyweight
        """
        return self.cached_createtimes.create_flyweight(object_name)

    def __enter__(self):
        self.cached_createtimes.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.cached_createtimes.__exit__(exc_type, exc_value, traceback)

    def __getstate__(self):
        return self.cached_createtimes

    def __setstate__(self, state):
        self.cached_createtimes = state

    def create_store(self, filename, **kwargs):
        """
        create a storage object for the file
        :param filename: name of object, should start with bucket name
        :type filename: str
        :param kwargs: keyword args for storage
        :type kwargs: dict
        :return: storage object
        :rtype: pypeliner.contrib.aws.objectstorage.AwsObjectStorage
        """
        return AwsObjectStorage(self, filename, **kwargs)

    def unpack_path(self, filename):
        """
        extract bucket and object from path
        :param filename: file path
        :type filename: str
        :return: bucket and object names
        :rtype: list of str
        """
        if filename.startswith('/'):
            filename = filename[1:]
        filename = filename.split('/')
        bucket = filename[0]
        filename = '/'.join(filename[1:])
        return bucket, filename

    def push(self, object_name, filename, createtime):
        """
        push data to s3
        :param object_name: name of object, should start with bucket name
        :type object_name: str
        :param filename: path on local storage
        :type filename: str
        :param createtime: time string
        :type createtime: str
        """
        bucket, object_name = self.unpack_path(object_name)
        self.connect()
        self.object_client.create_bucket(bucket)
        self.object_client.upload_from_file(
            filename, object_name, bucket)
        self.object_client.set_metadata(bucket, object_name, 'create_time', createtime)

    def pull(self, object_name, filename):
        """
        pull data from S3
        :param object_name: name of object, should start with bucket name
        :type object_name: str
        :param filename: path on local storage
        :type filename: str
        """
        bucket, object_name = self.unpack_path(object_name)
        self.connect()
        self.object_client.download_to_path(object_name, bucket, filename)
        object_size = self.object_client.get_file_size(bucket, object_name)
        file_size = os.path.getsize(filename)
        if file_size != object_size:
            raise Exception('file size mismatch for {}:{} compared to {}:{}'.format(
                object_name, object_size, filename, file_size))

    def retrieve_object_createtime(self, object_name):
        """
        get the create time from object metadata
        :param object_name: name of object, should start with bucket name
        :type object_name: str
        :return: time str
        :rtype: str
        """
        bucket, object_name = self.unpack_path(object_name)
        self.connect()
        createtime = self.object_client.get_metadata(bucket, object_name, 'create_time')
        if not createtime:
            createtime = self.object_client.get_last_modified(bucket, object_name)
        return createtime

    def update_object_createtime(self, object_name, createtime):
        """
        update createtime in S3 metadata
        :param object_name: name of object, should start with bucket name
        :type object_name: str
        :param createtime: time string
        :type createtime: str
        """
        bucket, object_name = self.unpack_path(object_name)
        self.connect()
        self.object_client.set_metadata(
            bucket,
            object_name,
            'create_time',
            createtime)

    def delete_object(self, object_name):
        """
        delete an object from S3
        :param object_name: name of object, should start with bucket name
        :type object_name: str
        """
        bucket, object_name = self.unpack_path(object_name)
        self.connect()
        self.object_client.delete_key(
            bucket,
            object_name)



from __future__ import absolute_import, print_function

import os
import dill as pickle
import logging
import yaml
import time
import pypeliner.execqueue.base
import pypeliner.tests.jobs
import pypeliner.helpers
import pypeliner.contrib.aws.helpers as aws_helpers

from .objectstorage import AwsSimpleStorageService

from .aws_batch import AwsBatch

class AwsJobQueue(object):
    """ Azure batch job queue.
    """
    debug_filenames = {
        'job stderr': 'stderr.txt',
        'job stdout': 'stdout.txt',
    }

    def __init__(self, config_filename=None, **kwargs):
        aws_helpers.set_aws_logging_filters()

        self.run_id = aws_helpers.random_string(8)

        with open(config_filename) as f:
            self.config = yaml.safe_load(f)

        self.logger = logging.getLogger('pypeliner.execqueue.aws_batch')

        self.logger.info('creating blob client')

        self.s3client = AwsSimpleStorageService()

        self.batch_client = AwsBatch()

        self.logger.info('creating task container')

        self.job_storage_bucket = self.config['storage_bucket']
        self.s3client.create_bucket(self.job_storage_bucket)

        self.job_names = {}
        self.job_ids = {}
        self.job_temps_dir = {}

        self.job_blobname_prefix = {}
        self.running_job_ids = set()
        self.completed_job_ids = set()
        self.job_names = {}
        self.job_definitions = {}

    def __enter__(self):
        """
        Create the compute environments and job queues if needed
        """
        self.batch_client.create_compute_environments(self.config)
        # need to add a little bit of delay, to let AWS create env
        time.sleep(5)
        self.batch_client.create_job_queues(self.config, self.run_id)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        """
        kill all running jobs
        """
        reason = '{}:{}'.format(str(exc_type), str(exc_value))
        for job_id in self.running_job_ids:
            self.batch_client.terminate_job(job_id, reason)

        self.batch_client.delete_job_queues(self.config, self.run_id)
        self.batch_client.delete_compute_environments(self.config)

    def _create_error_text(self, job_temp_dir, hostname=None):
        """
        create error if the delegator fails and doesnt generate
        after pickle
        :param job_temp_dir: temp dir used for saving before and after pickles
        :type job_temp_dir: str
        :param hostname: machine hostname where the job was executed
        :type hostname: str
        :return: error explanation
        :rtype: str
        """
        error_text = 'failed to execute'

        if hostname:
            error_text += ' on node {}'.format(hostname)

        error_text = [error_text]

        for debug_type, debug_filename in self.debug_filenames.items():
            debug_filename = os.path.join(job_temp_dir, debug_filename)

            if not os.path.exists(debug_filename):
                error_text += [debug_type + ': missing']

                continue

            with open(debug_filename, 'r') as debug_file:
                error_text += [debug_type + ':']

                for line in debug_file:
                    error_text += ['\t' + line.rstrip()]

        return '\n'.join(error_text)

    def send(self, ctx, name, sent, temps_dir):
        """
        Add a job to the queue.
        :param ctx: context of job, mem etc.
        :type ctx: dict
        :param name: unique name for the job
        :type name: str
        :param sent: callable object
        :type sent: callable
        :param temps_dir: unique path for job temps
        :type temps_dir: str
        """
        self.job_temps_dir[name] = temps_dir

        job_name = self.batch_client.batch_compatible_format(name)
        # sanity check, remove this later
        jobs = [job['jobName'] for job in self.batch_client.list_all_jobs(self.config, self.run_id)]
        if job_name in jobs:
            raise Exception()

        self.job_blobname_prefix[name] = 'output_' + job_name

        # create pickle
        job_before_filename = os.path.join(temps_dir, 'job.pickle')
        with open(job_before_filename, 'wb') as before:
            pickle.dump(sent, before)

        # upload pickle to s3
        job_before_blobname = os.path.join(
            self.job_blobname_prefix[name],
            'job.pickle')
        self.s3client.upload_from_file(
            job_before_filename, job_before_blobname, self.job_storage_bucket
        )

        workdir = os.path.join(self.config['workdir'], job_name)
        command = ['aws_fetch_run', self.job_storage_bucket, job_before_blobname, workdir]

        context_config = pypeliner.helpers.GlobalState.get("context_config")

        defname = self.job_definitions.get(ctx.get("image"))
        if not defname:
            defname = self.batch_client.get_matching_job_defn(self.config, ctx, context_config)
            self.job_definitions[ctx.get("image")] = defname

        self.logger.debug('job {} submitted with job name: {}'.format(name, job_name))

        job_id = self.batch_client.submit_job(job_name, defname, ctx, command, self.config, self.run_id)

        self.running_job_ids.add(job_id)
        self.job_names[job_id] = name
        self.job_ids[name] = job_id

    def wait(self, immediate=False):
        """
        Wait for a job to finish.
        :param immediate: do not wait if no job has finished
        :type immediate: bool
        :return: job name
        :rtype: str
        """
        while True:
            jobs = self.batch_client.list_all_jobs(
                self.config, self.run_id, status=["FAILED", "SUCCEEDED"]
            )

            for jobsummary in jobs:
                jobid = jobsummary['jobId']
                self.completed_job_ids.add(jobid)

            for job_id in self.running_job_ids.intersection(
                    self.completed_job_ids):
                return self.job_names[job_id]

    def receive(self, name):
        """
        receive finished job
        :param name: name of the job to retrieve
        :type name: str
        :return: received object
        :rtype: object
        """
        blobname_prefix = self.job_blobname_prefix.pop(name)
        job_id = self.job_ids.pop(name)
        self.job_names.pop(job_id)
        temps_dir = self.job_temps_dir.pop(name)
        self.running_job_ids.remove(job_id)

        # check aws batch for exitcode
        exitcode = self.batch_client.get_job_exitcode(job_id)
        if not exitcode == 0:
            cloudwatch_log = self.batch_client.get_cloudwatch_error(job_id)
            if cloudwatch_log:
                raise pypeliner.execqueue.base.ReceiveError(
                    cloudwatch_log)
            else:
                raise pypeliner.execqueue.base.ReceiveError(
                    self._create_error_text(temps_dir))

        job_after_filename = os.path.join(temps_dir, 'job.pickle.after')

        try:
            self.s3client.download_to_path(
                os.path.join(blobname_prefix, 'job.pickle.after'),
                self.config['storage_bucket'],
                os.path.join(temps_dir, 'job.pickle.after'))
        except:
            raise pypeliner.execqueue.base.ReceiveError(
                self._create_error_text(temps_dir))

        with open(job_after_filename, 'rb') as job_after_file:
            received = pickle.load(job_after_file)

        if received is None or not received.started:
            raise pypeliner.execqueue.base.ReceiveError(
                self._create_error_text(temps_dir)
            )

        received.hostname = 'aws_ecs'

        return received

    @property
    def length(self):
        """
        :return: Number of jobs in the queue.
        :rtype: int
        """
        return len(self.job_names)

    @property
    def empty(self):
        """
        :return: True if queue is empty
        :rtype: bool
        """
        return self.length == 0


if __name__ == "__main__":
    exc_dir = 'tempazure'

    logging.basicConfig(level=logging.DEBUG)

    with AwsJobQueue() as queue:
        job = pypeliner.tests.jobs.TestJob()
        pypeliner.helpers.makedirs(exc_dir)

        queue.send({}, 'testjob1', job, exc_dir)

        finished_name = queue.wait()

        job2 = queue.receive(finished_name)



import os
import boto3
import time


class AwsBatch(object):
    def __init__(self):
        self.batch_client = boto3.client('batch')
        self.log_client = boto3.client('logs')

    def register_job_definition(self, defn_name, container_prop, retry_strategy, container_type):
        """
        register a job definition with aws batch
        :param defn_name: name for the job def
        :type defn_name: str
        :param container_prop: container properties for job def
        :type container_prop: dict
        :param retry_strategy: number of retries allowed
        :type retry_strategy: dict
        :param container_type: type of container (container|Multinode)
        :type container_type: str
        :return: ARN for the registered job def
        :rtype: str
        """
        response = self.batch_client.register_job_definition(
            jobDefinitionName=defn_name,
            type=container_type,
            containerProperties=container_prop,
            retryStrategy=retry_strategy,
        )
        return response['jobDefinitionArn']

    def batch_compatible_format(self, string):
        """
        for string to make it compatible with AWS batch naming conventions
        :param string: input string
        :type string: str
        :return: formatted string
        :rtype: str
        """
        string = string.replace('/', '_')
        string = string.replace(':', '_')
        string = string.replace('.', '')

        if string.startswith('_'):
            string = string[1:]
        return string

    def _compare_dicts(self, ref1, ref2):
        """
        compares 2 dictionaries
        :param ref1: first dict to compare
        :type ref1: dict
        :param ref2: second dict to compare
        :type ref2: dict
        :return: True if the dicts are same
        :rtype: bool
        """
        keys = set(list(ref1.keys()) + list(ref2.keys()))

        for key in keys:
            ref1val = ref1.get(key, None)
            ref2val = ref2.get(key, None)

            # also ignores empty lists. for instance if ulimits are not specified
            # when registering a job then aws batch returns [] for those defns.
            if not ref1val and not ref2val:
                continue

            if isinstance(ref1val, dict) and isinstance(ref2val, dict):
                result = self._compare_dicts(ref1val, ref2val)
                if not result:
                    return False

            if not ref1val == ref2val:
                return False

        return True

    @staticmethod
    def _get_boto3_iterator(func, kwargs, response_key):
        """
        utility function to get an iterator over batch list functions
        :param func: batch function to iterate over
        :type func: function
        :param kwargs: arguments for the func
        :type kwargs: dict
        :param response_key: key-value pair in aws response to iterate over
        :type response_key: str
        :return: values for the response key
        :rtype: generator
        """
        def run_func(next_token):
            if next_token:
                return func(**dict(nextToken=next_token, **kwargs))
            else:
                return func(**kwargs)

        next_token=None
        while True:
            response = run_func(next_token)

            next_token = response.get('nextToken', None)

            for value in response[response_key]:
                yield value
            if not next_token:
                break

    def _search_matching_job_defn(self, retry_strategy, container_properties, container_type):
        """
        look through all revisions of existing job definitions and return
        the revision that matches the request.
        :param retry_strategy: requested retry strategy
        :type retry_strategy: dict
        :param container_properties: requested conatiner properties
        :type container_properties: dict
        :param container_type: type (container | multinode)
        :type container_type: str
        :return: matching job definition
        :rtype: str
        """
        job_defns = self._get_boto3_iterator(
            self.batch_client.describe_job_definitions,
            {'maxResults': 10, 'status': 'ACTIVE'},
            'jobDefinitions'
        )

        for job_defn in job_defns:
            defn_container_props = job_defn['containerProperties']
            defn_retry = job_defn.get('retryStrategy',0)
            defn_container_type = job_defn['type']
            defn_arn = job_defn['jobDefinitionArn']

            match = True
            if not defn_retry == retry_strategy:
                match = False
            if not self._compare_dicts(defn_container_props, container_properties):
                match = False
            if not defn_container_type == container_type:
                match = False
            if match:
                return defn_arn
        return

    def get_matching_job_defn(self, config, ctx, context_config):
        """
        build kwargs to request new job defn, check existing revisions
        to see if there is matching one already. if not then create a new one
        :param config: AWS batch submit config
        :type config: dict
        :param ctx: job context
        :type ctx: dict
        :param context_config: pypeliner context config
        :type context_config: dict
        :return: job definition ARN
        :rtype:  str
        """
        image_name = ctx.get("docker_image")
        image = os.path.join(context_config['docker']['server'], image_name)
        defname = self.batch_compatible_format(image_name)
        def_metadata = config['definition']

        mount_config = context_config['docker']['mounts']
        # required for docker inside docker
        mount_config['docker_socket'] = '/var/run/docker.sock'
        mount_config['docker'] = '/usr/bin/docker'

        volumes = []
        for name, path in mount_config.items():
            volumes.append({'name': name, 'host': {'sourcePath': path}})

        mountpoints = []
        for name, path in mount_config.items():
            mountpoints.append({'containerPath': path, 'sourceVolume': name, 'readOnly': False})

        env_vars = []
        if def_metadata['env_vars']:
            for name, value in def_metadata['env_vars'].items():
                env_vars.append({'name': name, 'value': value})

        container_properties = {
            'image': image,
            'vcpus': def_metadata['vcpus'],
            'memory': def_metadata['mem'],
            'command': def_metadata['command'],
            'jobRoleArn': def_metadata['job_role_arn'],
            'volumes': volumes,
            'environment': env_vars,
            'mountPoints': mountpoints,
            'readonlyRootFilesystem': def_metadata['readonlyRootFilesystem'],
            'privileged': def_metadata['privileged'],
        }

        retry_strategy = {'attempts': 1}

        matching_job_def = self._search_matching_job_defn(
            container_properties, retry_strategy, def_metadata['type']
        )
        if not matching_job_def:
            matching_job_def = self.register_job_definition(
                defname, container_properties,
                retry_strategy, def_metadata['type']
            )
        return matching_job_def

    def list_all_jobs(self, config, run_id, status=None):
        """
        lists all jobs for specified run id
        :param config: pypeliner batch submit config
        :type config: dict
        :param run_id: pipeline run id
        :type run_id: str
        :return: summaries of all jobs
        :rtype: list
        """
        queues = []
        config = config.get("job_queues")
        spot_queue = config["spot"]
        queues.append('{}_{}'.format(spot_queue['queue_name_prefix'], run_id))
        dedicated_queue = config["dedicated"]
        queues.append('{}_{}'.format(dedicated_queue['queue_name_prefix'], run_id))
        return self.list_jobs(queues, status=status)


    def list_jobs(self, job_queues, status=None):
        """
        list all jobs
        :param job_queues: job queues to look for jobs in
        :type job_queues: list of str
        :param status: only list jobs with this status
        :type status: list of str
        :return: job summaries for matching jobs
        :rtype: generator
        """
        valid_statuses = ['SUBMITTED', 'PENDING', 'RUNNABLE', 'STARTING',
                          'RUNNING', 'SUCCEEDED', 'FAILED']
        if not status:
            status = valid_statuses

        for stat in status:

            for job_queue in job_queues:

                assert stat in valid_statuses

                jobs = self._get_boto3_iterator(
                    self.batch_client.list_jobs,
                    {'jobQueue': job_queue, 'jobStatus': stat, 'maxResults': 10},
                    'jobSummaryList'
                )

                for job in jobs:
                    yield job

    def get_compute_environment_status(self, env_name):
        """
        get status of compute environments
        :return: names of all compute environments
        :rtype: list of str
        """
        env = self._get_boto3_iterator(
            self.batch_client.describe_compute_environments,
            {'computeEnvironments': [env_name]},
            'computeEnvironments'
        )
        env = list(env)
        assert len(env) == 1
        env = env[0]
        return env.get("status")

    def get_job_queue_status(self, queue_name):
        """
        get status of compute environments
        :return: names of all compute environments
        :rtype: list of str
        """
        queue = self._get_boto3_iterator(
            self.batch_client.describe_job_queues,
            {'jobQueues': [queue_name]},
            'jobQueues'
        )
        queue = list(queue)
        assert len(queue) == 1
        queue = queue[0]
        return queue.get("status")

    def list_compute_environments(self):
        """
        list all existing compute environments
        :return: names of all compute environments
        :rtype: list of str
        """
        envs = self._get_boto3_iterator(
            self.batch_client.describe_compute_environments,
            {'maxResults': 10},
            'computeEnvironments'
        )
        envs = [env['computeEnvironmentName'] for env in envs]
        return envs

    def list_job_queues(self):
        """
        lists all job queues
        :return: names of all job queues
        :rtype: list of str
        """
        envs = self._get_boto3_iterator(
            self.batch_client.describe_job_queues,
            {'maxResults': 10},
            'jobQueues'
        )
        envs = [env['jobQueueName'] for env in envs]
        return envs

    def create_compute_environment(self, env_name, env_params):
        """
        create a new compute environment
        :param env_name: name of env
        :type env_name: str
        :param env_params: params from submit config
        :type env_params: dict
        """
        compute_resources = {
           'type': env_params['type'],
           'minvCpus': env_params['minvCpus'],
           'maxvCpus': env_params['maxvCpus'],
           'desiredvCpus': env_params['desiredvCpus'],
           'instanceTypes': env_params['instanceTypes'],
           'imageId': env_params['imageId'],
           'subnets': env_params['subnets'],
           'securityGroupIds': env_params['securityGroupIds'],
           'ec2KeyPair': env_params['ec2KeyPair'],
           'instanceRole': env_params['instanceRole'],
           'tags': env_params['tags'],
           'placementGroup': env_params['placementGroup'],
           'bidPercentage': env_params['bidPercentage'],
           'spotIamFleetRole': env_params['spotIamFleetRole'],
           'launchTemplate': env_params['launchTemplate']
        }

        # boto3 doesnt accept None
        pop_keys = [k for k, v in compute_resources.items() if v is None]
        [compute_resources.pop(k) for k in pop_keys]

        response = self.batch_client.create_compute_environment(
            computeEnvironmentName=env_name,
            type='MANAGED',
            state='ENABLED',
            computeResources=compute_resources,
            serviceRole=env_params['serviceRole']
        )

    def create_compute_environments(self, config):
        """
        create all compute environments in submit config
        if required
        :param config: aws batch submit configuration data
        :type config: dict
        """
        existing_envs = self.list_compute_environments()

        config = config.get("compute_environments")
        if not config:
            return

        for env_name, env_params in config.items():
            if env_name in existing_envs:
                continue
            self.create_compute_environment(env_name, env_params)

    def create_job_queue(self, queue_name, queue_params):
        """
        create a job queue
        :param queue_name:  name of the queue
        :type queue_name: str
        :param queue_params: queue params from submit config
        :type queue_params: dict
        """
        response = self.batch_client.create_job_queue(
            jobQueueName=queue_name,
            state='ENABLED',
            priority=queue_params['priority'],
            computeEnvironmentOrder=[
                {
                    'order': 1,
                    'computeEnvironment': queue_params['computeEnvironment']
                },
            ]
        )

    def create_job_queues(self, config, run_id):
        """
        create all job queues in submit config
        if required
        :param run_id: pipeline run id
        :type run_id: str
        :param config: aws batch submit configuration data
        :type config: dict
        """
        config = config.get("job_queues")
        existing_queues = self.list_job_queues()

        spot_queue = config["spot"]
        queue_name = '{}_{}'.format(spot_queue['queue_name_prefix'], run_id)
        if queue_name not in existing_queues:
            self.create_job_queue(queue_name, spot_queue)

        dedicated_queue = config["dedicated"]
        queue_name = '{}_{}'.format(dedicated_queue['queue_name_prefix'], run_id)
        if queue_name not in existing_queues:
            self.create_job_queue(queue_name, dedicated_queue)

    def get_log_stream_name(self, job_id):
        """
        get the logstream name for the job.
        :param job_id: job Id
        :type job_id: str
        :return: log stream name
        :rtype:  str
        """
        desc = self.batch_client.describe_jobs(jobs=[job_id])
        assert len(desc['jobs']) == 1
        desc = desc['jobs'][0]
        return desc['container']['logStreamName']

    def get_cloudwatch_error(self, job_id):
        """
        fetch all logs for the job
        :param job_id: job Id
        :type job_id: str
        :return: logs for the job
        :rtype: str
        """
        logstream = self.get_log_stream_name(job_id)

        kwargs = {
            'logGroupName': '/aws/batch/job',
            'logStreamName': logstream,
            'limit': 100,
            'startFromHead': True
        }


        logs = self._get_boto3_iterator(
            self.log_client.get_log_events,
            kwargs,
            'events'
        )

        data = [event.message for event in logs]

        return '\n'.join(data)

    def get_job_exitcode(self, job_id):
        """
        get exit code from aws batch
        :param job_id: job Id
        :type job_id: str
        :return: exit code
        :rtype: int
        """
        desc = self.batch_client.describe_jobs(jobs=[job_id])

        assert len(desc['jobs']) == 1

        status = desc['jobs'][0]['status']

        assert status in ['SUCCEEDED', 'FAILED'], "job is not in finished state"

        attempts = desc['jobs'][0]['attempts']

        attempt_codes = [attempt['container']['exitCode'] for attempt in attempts]

        # exit code should be 0 if job succeeded.
        if not status == 'FAILED':
            assert list(set(attempt_codes)) == [0]

        # only reporting the last exit code if multiple attempts
        exitcode = attempt_codes[-1]
        return exitcode

    def terminate_job(self, job_id, reason):
        """
        kill a job
        :param job_id: job Id
        :type job_id: str
        :param reason: reason for killing
        :type reason: str
        """
        self.batch_client.terminate_job(jobId=job_id, reason=reason)

    def submit_job(self, job_name, defn_id, ctx, command, config, run_id):
        """
        submit a job
        :param job_name: unique job name
        :type job_name: str
        :param defn_id: job definition name or ARN
        :type defn_id: str
        :param ctx: job context
        :type ctx: dict
        :param command: job command
        :type command:  list
        :param config: batch submit config
        :param run_id: pipeline run id
        :type run_id: str
        :type config: dict
        :return: job Id
        :rtype: str
        """
        if ctx.get("dedicated"):
            jobqueue = config['job_queues']['dedicated']['queue_name_prefix']
        else:
            jobqueue = config['job_queues']['spot']['queue_name_prefix']
        jobqueue = "{}_{}".format(jobqueue, run_id)
        response = self.batch_client.submit_job(
            jobName=job_name,
            jobQueue=jobqueue,
            jobDefinition=defn_id,
            containerOverrides={
                'vcpus': ctx.get('ncpus'),
                'memory': ctx.get('mem') * 1024,
                'command': command,
            }
        )

        job_id = response['jobId']

        return job_id

    def delete_job_queue(self, job_queue):
        """
        delete a job queue
        :param job_queue: job queue name
        :type job_queue: str
        """
        self.batch_client.delete_job_queue(
            jobQueue=job_queue
        )

        while job_queue in self.list_job_queues():
            time.sleep(10)

    def disable_job_queue(self, job_queue):
        """
        disable a job queue
        :param job_queue: job queue name
        :type job_queue: str
        """

        self.batch_client.update_job_queue(
            jobQueue=job_queue,
            state='DISABLED'
        )

        while True:
            status = self.get_job_queue_status(job_queue)
            if not status == 'UPDATING':
                break
            time.sleep(10)

    def delete_compute_environment(self, compute_environment):
        """
        delete a compute environment
        :param compute_environment: compute environment name
        :type compute_environment: str
        """
        self.batch_client.delete_compute_environment(
            computeEnvironment=compute_environment
        )

        while compute_environment in self.list_compute_environments():
            time.sleep(10)

    def disable_compute_environment(self, compute_environment):
        """
        disable a compute environment
        :param compute_environment: compute environment name
        :type compute_environment: str
        """
        self.batch_client.update_compute_environment(
            computeEnvironment=compute_environment,
            state='DISABLED'
        )

        while True:
            status = self.get_compute_environment_status(compute_environment)
            if not status == 'UPDATING':
                break
            time.sleep(10)

    def delete_job_queues(self, config, run_id):
        """
        delete all job queues
        :param config: pypeliner submit config
        :type config: dict
        :param run_id: pipeline run id
        :type run_id: str
        """
        if not config.get('delete_job_queues'):
            return

        config = config.get("job_queues")

        spot_queue = config["spot"]
        queue_name = '{}_{}'.format(spot_queue['queue_name_prefix'], run_id)
        self.disable_job_queue(queue_name)
        self.delete_job_queue(queue_name)

        dedicated_queue = config["dedicated"]
        queue_name = '{}_{}'.format(dedicated_queue['queue_name_prefix'], run_id)
        self.disable_job_queue(queue_name)
        self.delete_job_queue(queue_name)

    def delete_compute_environments(self, config):
        """
        delete all compute environments
        :param config: pypeliner submit config
        :type config: dict
        """
        if not config.get('delete_compute_environments'):
            return

        config = config.get("compute_environments")

        for env_name, env_params in config.items():
            self.disable_compute_environment(env_name)
            self.delete_compute_environment(env_name)




import importlib
import pypeliner.execqueue.local


def create(requested_queue, modules=None, native_spec=None, config_filename=None):
    if requested_queue is None:
        raise Exception('No submit queue specified')
    elif requested_queue == 'local':
        exec_queue_name = 'pypeliner.execqueue.local.LocalJobQueue'
    elif requested_queue == 'qsub':
        exec_queue_name = 'pypeliner.execqueue.qsub.QsubJobQueue'
    elif requested_queue == 'asyncqsub':
        exec_queue_name = 'pypeliner.execqueue.qsub.AsyncQsubJobQueue'
    elif requested_queue == 'lsf':
        exec_queue_name = 'pypeliner.execqueue.qsub.LsfJobQueue'
    elif requested_queue == 'slurm':
        exec_queue_name = 'pypeliner.execqueue.qsub.SlurmJobQueue'
    elif requested_queue == 'pbs':
        exec_queue_name = 'pypeliner.execqueue.qsub.PbsJobQueue'
    elif requested_queue == 'drmaa':
        exec_queue_name = 'pypeliner.execqueue.drmaa.DrmaaJobQueue'
    elif requested_queue == 'azurebatch':
        exec_queue_name = 'pypeliner.contrib.azure.batchqueue.AzureJobQueue'
    elif requested_queue == 'awsbatch':
        exec_queue_name = 'pypeliner.contrib.aws.batchqueue.AwsJobQueue'
    else:
        exec_queue_name = requested_queue

    exec_queue_class_name = exec_queue_name.split('.')[-1]
    exec_queue_module_name = exec_queue_name[:-len(exec_queue_class_name)-1]

    exec_queue_module = importlib.import_module(exec_queue_module_name)
    exec_queue_class = vars(exec_queue_module)[exec_queue_class_name]

    exec_queue = exec_queue_class(modules=modules, native_spec=native_spec, config_filename=config_filename)

    if requested_queue != 'local':
        exec_queue = pypeliner.execqueue.local.LocalRemoteQueue(exec_queue, modules=modules)

    return exec_queue



import os


def log_text(debug_filenames):
    text = ''
    for debug_type, debug_filename in debug_filenames.items():
        preamble = '-' * 10 + ' ' + debug_type + ' ' + '-' * 10
        if not os.path.exists(debug_filename):
            text += preamble + ' (missing)\n'
            continue
        else:
            text += preamble + '\n'
        with open(debug_filename, 'r') as debug_file:
            text += debug_file.read()
    return text


def qsub_format_name(name):
    return name.strip('/').rstrip('/').replace('/', '.').replace(':', '_')
